/*
 * Dfmpr:  Derivative free minimization palomares rodriguez
 *
 * Copyright (c) 2002-2013
 *   Arno Formella (formella@ei.uvigo.es)
 *   Ubaldo Garcia-Palomares (garciap@usb.ve)
 *   Jose F. Rodriguez (jrodri@usb.ve)
 */

/*
 * Academic Free License v2.0
 * 
 * http://www.opensource.org/licenses/
 */

template<typename V>
Dfmpr<V>::Dfmpr(
  OptimizationProblem<V>& problem_,
  unsigned int max_iterations_,
  unsigned int max_evaluations_,
  double tolerance_,
  double tau_,
  double delta_magic,
  double expansion_constant_,
  double contraction_constant_
) :
  Optimize<V>(
    problem_,
    max_iterations_,
    max_evaluations_,
    tolerance_
  ),
  tau(tau_<0.001?0.001:tau_),
  delta(
    ::fabs(delta_magic)>1.0?
      1.0/sqrt((double)Optimize<V>::dimension):
      ::fabs(delta_magic)/sqrt((double)Optimize<V>::dimension)
  ),
  expansion_constant(
    ::fabs(expansion_constant_)<=1.0?1.8:
      ::fabs(expansion_constant_)
  ),
  contraction_constant(
    ::fabs(contraction_constant_)>=1.0?0.2:
      ::fabs(contraction_constant_)
  )
{
}

template<typename V>
Dfmpr<V>::~Dfmpr(
) {
}

template<typename V>
void Dfmpr<V>::CheckLimits(
  V* inf,
  V* sup
) const {
  for(int i(Optimize<V>::dimension); i>0; ) {
    --i;
    if((*inf)[i]>(*sup)[i]) {
      double h((*sup)[i]);
      (*sup)[i]=(*inf)[i];
      (*inf)[i]=h;
    }
  }
}

template<typename V>
double Dfmpr<V>::Median(
  double a,
  double b,
  double c
) const {
  if(a<b) {
    if(b<c) return b;
    if(a<c) return c;
    return a;
  }
  if(a<c) return a;
  if(c<b) return b;
  return c;
}

template<typename V>
void Dfmpr<V>::Median(
  const V& a,
  const V& b,
  const V& c,
  V& x
) const {
  for(int i(Optimize<V>::dimension); i>0; ) {
    --i;
    x[i]=Median(a[i],b[i],c[i]);
  }
}

template<typename V>
typename Optimize<V>::Optresult Dfmpr<V>::Run(
  V& v,
  V* inf,
  V* sup
) const {
  Optimize<V>::evaluations=0;
  Optimize<V>::iterations=0;
  Optimize<V>::minimum_fobj=std::numeric_limits<double>::max();

  bool have_limits(inf!=0&&sup!=0);
  if(have_limits) CheckLimits(inf,sup);
  unsigned int iter(0);

  V x(v);
  if(have_limits) Median(*inf,v,*sup,v);
  double fmin(Optimize<V>::problem.Fobj(x));
  unsigned int val(1);

  V h(Optimize<V>::dimension);
  double tau_local(tau);
  for(unsigned int i(Optimize<V>::dimension);i>0;) h[--i]=tau_local;

  V xold(x);
  V u(Optimize<V>::dimension);
  V z(Optimize<V>::dimension);
  V d(Optimize<V>::dimension);

  bool success(false);

  const double tol(mcl::Optimize<V>::tolerance);

  while(
    tau_local>tol&&
    val<Optimize<V>::max_evaluations&&
    iter<Optimize<V>::max_iterations
  ) {
    if(success) {
      u=x-xold;
      success=false;
    }
    else {
      for(unsigned int i(Optimize<V>::dimension);i>0;)
        u[--i]=-1.0+2.0*Rand();
    }
    u=u.norm();

    bool any_change(false);
    for(unsigned int k(Optimize<V>::dimension); k>0;) {
      if(::fabs(u[--k])-delta<0.0) {
        any_change=true;
        if(have_limits)
          u[k]=u[k]-(*inf)[k]<(*sup)[k]-u[k]?1.0:-1.0;
        else
          u[k]=delta*(u[k]<0.0?-1.0:1.0);
      }
    }
    if(any_change) u=u.norm();

    xold=x;
   
    for(unsigned int i(Optimize<V>::dimension);i>0;) h[--i]=tau_local;
    double hmin(0.0);
    int imin(-1);
    int index(-1);
    unsigned int faults(0);
    while(faults<Optimize<V>::dimension+Optimize<V>::dimension) {
      index++;
      if(index>=(int)Optimize<V>::dimension) index=0;

      GenerateD(u,index,d);
      
      while(1) {
        z=x+h[index]*d;
        if(have_limits) Median(*inf,z,*sup,z);
        double fz(Optimize<V>::problem.Fobj(z));
        val++;
        if(fz<fmin-tau_local*tau_local) {
          x=z;
          fmin=fz;
          imin=-1;
          success=true;
          faults=0;
          h[index]*=expansion_constant;
        }
        else {
          faults++;
          if(fz<fmin) {
            fmin=fz;
            imin=index;
            hmin=h[index];
          }
          break;
        }
      }
      if(faults==Optimize<V>::dimension) h=-h;     
    }
   
    iter++;
    if(imin>=0) {
      GenerateD(u,imin,d);
      x=x+hmin*d;
      if(have_limits) Median(*inf,x,*sup,x);
    }
   
    h=contraction_constant*h;
    double max_h(std::numeric_limits<double>::min());
    double min_h(std::numeric_limits<double>::max());
    for(unsigned int i(Optimize<V>::dimension);i>0;) {
      --i;
      h[i]=::fabs(h[i]);
      if(max_h<h[i]) max_h=h[i];
      if(min_h>h[i]) min_h=h[i];
    }
    tau_local=Median(tau_local,min_h,max_h);
  }
  Optimize<V>::evaluations=val;
  Optimize<V>::iterations=iter;
  Optimize<V>::minimum_fobj=fmin;
  v=x;
  if(iter>=Optimize<V>::max_iterations) return Optimize<V>::max_iter_exceeded;
  if(val>=Optimize<V>::max_evaluations) return Optimize<V>::max_eval_exceeded;
  return Optimize<V>::converged;
}

template<typename V>
void Dfmpr<V>::GenerateD(
  const V& u,
  int index,
  V& d
) const {
  d=-2.0*u[index]*u;
  d[index]+=1.0;
  if(u[index]>0.0) d=-d;
}

template<typename V>
double Dfmpr<V>::Rand(
) const {
#ifdef HAVE_MCL_RANDOM
  return R.GetDouble();
#else
  return rand()/static_cast<double>(RAND_MAX);
#endif
}

template<typename V>
void Dfmpr<V>::SetRandomSeed(
) const {
#ifdef HAVE_MCL_RANDOM
  R.SetRandomSeed();
#else
  // a little bit strange, but ...
  srand(static_cast<unsigned int>(time(0)));
#endif
}

template<typename V>
void Dfmpr<V>::SetSeed(
  unsigned int seed
) const {
#ifdef HAVE_MCL_RANDOM
  R.SetSeed(seed);
#else
  srand(seed);
#endif
}

