/*
 * Optimize:  abstract class for function minimization 
 *
 * Copyright (c) 2002-2013,
 *   Arno Formella (formella@ei.uvigo.es)
 *   Ubaldo Garcia-Palomares (garciap@usb.ve)
 *   Jose F. Rodriguez (jrodri@usb.ve)
 */

/*
 * Academic Free License v2.0
 * 
 * http://www.opensource.org/licenses/
 */

#ifndef __MCL_OPTIMIZE_H__
#define __MCL_OPTIMIZE_H__

#include <assert.h>
#include <limits>

namespace mcl {

template<typename V>
class OptimizationProblem {
public:
  virtual ~OptimizationProblem() {}
  virtual unsigned int Dimension(void) const = 0;
  /// THE objective function.
  virtual double Fobj(
    const V&  //!< parameter vector where to evaluate
  ) = 0;
};

template<typename V>
class Optimize {
public:
  /// the return value of the optimization procedure indicates:
  enum Optresult {
    error,             //!< an error occurred
    converged,         //!< solution found (convergence occurred)
    max_iter_exceeded, //!< number of iterations exceeded maximum permited
    max_eval_exceeded, //!< number of evaluations exceeded maximum permited
    stagnation         //!< optimization procedure stagnated
  };
protected:
  unsigned int dimension;       //!< dimension of the problem to be optimized
  unsigned int max_iterations;  //!< maximum number of iterations allowed
  unsigned int max_evaluations; //!< maximum number of Fobj evaluations allowed
    // number of iterations actually performed
  mutable unsigned int iterations; 
    // number of evaluations actually performed
  mutable unsigned int evaluations;
  mutable double minimum_fobj;  //!< minimum value of the objective function
  double tolerance;             //!< termination criterion for convergence check
  OptimizationProblem<V>& problem;
private:
  /// copy-construction is not allowed (up to now ...)
  Optimize(const Optimize& D);
  /// direct assignment is not allowed (up to now ...)
  Optimize& operator=(const Optimize& D);
public:
  Optimize(
    OptimizationProblem<V>& problem_,
    unsigned int max_iterations_=1000,
    unsigned int max_evaluations_=1000,
    double tolerance_=0.00001
  ) :
    dimension(problem_.Dimension()),
    max_iterations(max_iterations_),
    max_evaluations(max_evaluations_),
    iterations(0),
    evaluations(0),
    minimum_fobj(std::numeric_limits<double>::max()),
    tolerance(tolerance_<1E-10?1E-10:tolerance_),
    problem(problem_)
  {
    assert(dimension!=0);
  }
  virtual ~Optimize() {}
  virtual unsigned int Dimension(void) const { return dimension; }
  virtual unsigned int Evaluations(void) const { return evaluations; }
  virtual unsigned int Iterations(void) const { return iterations; }
  virtual unsigned int MaxEvaluations(void) const { return max_evaluations; }
  virtual unsigned int MaxIterations(void) const { return max_iterations; }
  virtual double MinimumFobj(void) const { return minimum_fobj; }
  /// Optimize must be implemented by the derived class.
  /// The vector v must have the dimension of the problem.
  /// Optimize should return the optimal point in v.
  /// Either none or both limits must be present.
  /// Optimize should set evaluations, iterations and minimum_fobj
  /// accordingly
  virtual typename Optimize<V>::Optresult Run(
    V& v,      //!< the starting point for the optimization
    V* inf=0,  //!< limit inferior for optimal point (if present)
    V* sup=0   //!< limit superior for optimal point (if present)
  ) const = 0;
  double Tolerance(void) const { return tolerance; }
};

}

#endif
